/* wait.c, created from wait.def. */
#line 33 "./wait.def"

#line 43 "./wait.def"

#include <config.h>

#include "../bashtypes.h"
#include <signal.h>

#if defined (HAVE_UNISTD_H)
#  include <unistd.h>
#endif

#include "../bashansi.h"

#include "../shell.h"
#include "../jobs.h"
#include "common.h"
#include "bashgetopt.h"

extern int interrupt_immediately;

/* Wait for the pid in LIST to stop or die.  If no arguments are given, then
   wait for all of the active background processes of the shell and return
   0.  If a list of pids or job specs are given, return the exit status of
   the last one waited for. */

#define WAIT_RETURN(s) do { run_unwind_frame ("wait_builtin"); return (s); } while (0)

int
wait_builtin (list)
     WORD_LIST *list;
{
  int status;

  if (no_options (list))
    return (EX_USAGE);
  if (list != loptend)
    list = loptend;

  begin_unwind_frame ("wait_builtin");
  unwind_protect_int (interrupt_immediately);
  interrupt_immediately++;

  /* We support jobs or pids.
     wait <pid-or-job> [pid-or-job ...] */

  /* But wait without any arguments means to wait for all of the shell's
     currently active background processes. */
  if (list == 0)
    {
      wait_for_background_pids ();
      WAIT_RETURN (EXECUTION_SUCCESS);
    }

  status = EXECUTION_SUCCESS;
  while (list)
    {
      pid_t pid;
      char *w;

      w = list->word->word;
      if (digit (*w))
	{
	  if (all_digits (w + 1))
	    {
	      pid = (pid_t)atoi (w);
	      status = wait_for_single_pid (pid);
	    }
	  else
	    {
	      builtin_error ("`%s' is not a pid or valid job spec", w);
	      WAIT_RETURN (EXECUTION_FAILURE);
	    }
	}
#if defined (JOB_CONTROL)
      else if (job_control && *w)
	/* Must be a job spec.  Check it out. */
	{
	  int job;
	  sigset_t set, oset;

	  BLOCK_CHILD (set, oset);
	  job = get_job_spec (list);

	  if (job < 0 || job >= job_slots || !jobs[job])
	    {
	      if (job != DUP_JOB)
		builtin_error ("%s: no such job", list->word->word);
	      UNBLOCK_CHILD (oset);
	      status = 127;	/* As per Posix.2, section 4.70.2 */
	      list = list->next;
	      continue;
	    }

	  /* Job spec used.  Wait for the last pid in the pipeline. */
	  UNBLOCK_CHILD (oset);
	  status = wait_for_job (job);
	}
      else if (job_control == 0 && *w == '%')
	{
	  /* can't use jobspecs as arguments if job control is not active. */
	  builtin_error ("job control not enabled");
	  status = EXECUTION_FAILURE;
	}
#endif /* JOB_CONTROL */
      else
	{
	  builtin_error ("`%s' is not a pid or valid job spec", w);
	  status = EXECUTION_FAILURE;
	}
      list = list->next;
    }

  WAIT_RETURN (status);
}
