//
// convert.c - Partition-map format conversions.  LITTLE-ENDIAN CONVERSION
// IS NO LONGER DONE... deprecated and removed (dplatt@tivo.com, 5/15/09)
//
// Written by Eryk Vershen (eryk@apple.com)
//
// See comments in convert.h
//

/*
 * Copyright 2009 TiVo Inc. All Rights Reserved. by Apple Computer, Inc.
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * APPLE COMPUTER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL APPLE COMPUTER BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

#include <stdio.h>

#ifdef __linux__
#include <endian.h>
#else
#ifdef NeXT
#import <unistd.h>
#import <bsd/machine/endian.h>
#else
#define LITTLE_ENDIAN 1234
#define BIG_ENDIAN 4321
#define BYTE_ORDER 4321
#endif
#endif

#include <ctype.h>
#include "partition_map.h"
#include "convert.h"


//
// Defines
//


//
// Types
//


//
// Global Constants
//


//
// Global Variables
//


int bflag; /* -b option forces writing of big partition format */

//
// Routines
//
int
convert_dpme(DPME *data, int to_cpu_form)
{
    if (to_cpu_form) {
        /*
         * Partition map entry has been read from disk.  If it's in
         * legacy format, convert to current format.
         */
        if (data->dpme_signature == DPME_SIGNATURE) {
            DPMELegacy legacy;
            /*            printf("Importing partition map entry from legacy format\n"); */
            memcpy(&legacy, data, sizeof legacy);
            data->dpme_signature = TIVO_SIGNATURE;
            data->dpme_map_entries = legacy.dpme_map_entries;
            data->dpme_pblock_start = legacy.dpme_pblock_start;
            data->dpme_pblocks = legacy.dpme_pblocks;
            data->dpme_lblock_start = legacy.dpme_lblock_start;
            data->dpme_lblocks = legacy.dpme_lblocks;
            data->dpme_flags = legacy.dpme_flags;
            data->dpme_boot_block = legacy.dpme_boot_block;
            data->dpme_boot_bytes = legacy.dpme_boot_bytes;
            data->dpme_load_addr = legacy.dpme_load_addr;
            data->dpme_load_addr_2 = legacy.dpme_load_addr_2;
            data->dpme_goto_addr = legacy.dpme_goto_addr;
            data->dpme_goto_addr_2 = legacy.dpme_goto_addr_2;
            data->dpme_checksum = legacy.dpme_checksum;
            memcpy(data->dpme_name, legacy.dpme_name, sizeof data->dpme_name);
            memcpy(data->dpme_type, legacy.dpme_type, sizeof data->dpme_type);
        } else {
            /*            printf("Found partition map entry in new format\n"); */
        }
    } else {
        /*
         * Always export the first partition table entry into legacy
         * format, as it's the partition table descriptor itself and the
         * kernel currently doesn't know how to use the new format for
         * this one partition.
         *
         * Other partitions are left in the new format if the -b option is
         * specified (e.g. for testing) or if they refer to blocks which
         * go past the 32-bit barrier.  Otherwise they're converted back to
         * legacy format.
         */
        if (data->dpme_signature == TIVO_SIGNATURE &&
            ( data->dpme_pblock_start == 1 ||
              (!bflag && data->dpme_pblock_start <= 0x00000000FFFFFFFFLL &&
               data->dpme_pblocks      <= 0x00000000FFFFFFFFLL &&
               data->dpme_lblock_start <= 0x00000000FFFFFFFFLL &&
               data->dpme_lblocks      <= 0x00000000FFFFFFFFLL))) {
            DPMELegacy legacy;
            /* printf("Exporting partition map entry to legacy format\n"); */
            legacy.dpme_signature = DPME_SIGNATURE;
            legacy.dpme_map_entries = data->dpme_map_entries;
            legacy.dpme_pblock_start = data->dpme_pblock_start;
            legacy.dpme_pblocks = data->dpme_pblocks;
            legacy.dpme_lblock_start = data->dpme_lblock_start;
            legacy.dpme_lblocks = data->dpme_lblocks;
            legacy.dpme_flags = data->dpme_flags;
            legacy.dpme_boot_block = data->dpme_boot_block;
            legacy.dpme_boot_bytes = data->dpme_boot_bytes;
            legacy.dpme_load_addr = data->dpme_load_addr;
            legacy.dpme_load_addr_2 = data->dpme_load_addr_2;
            legacy.dpme_goto_addr = data->dpme_goto_addr;
            legacy.dpme_goto_addr_2 = data->dpme_goto_addr_2;
            legacy.dpme_checksum = data->dpme_checksum;
            memcpy(legacy.dpme_name, data->dpme_name, sizeof legacy.dpme_name);
            memcpy(legacy.dpme_type, data->dpme_type, sizeof legacy.dpme_type);
            memcpy(data, &legacy, sizeof *data);
        } else {
            /* printf("Leaving partition map entry in new format\n"); */
        }
    }
    return 0;
}


int
convert_block0(Block0 *data, int to_cpu_form)
{
    if (to_cpu_form) {
        // When reading from disk... old versions of pdisk will have set
        // the 32-bit field (now called sbBlkCountLegacy) correctly, to a
        // non-zero value, and the value of the new 64-bit count field at
        // the end of the sector will be undefined.  In this case, use the
        // legacy value.
        if (data->sbBlkCountLegacy != 0) {
            data->sbBlkCount = data->sbBlkCountLegacy;
        }
    } else {
        // For most disks, write the block count out to both fields, so
        // that legacy pdisk implementations and the kernel still
        // recognize it.  If the count is too large to fit into
        // 32 bits, set the legacy field to zero and write only the
        // 64-bit count.
        if (data->sbBlkCount <= 0x00000000FFFFFFFFLL) {
            data->sbBlkCountLegacy = data->sbBlkCount;
        } else {
            data->sbBlkCountLegacy = 0;
        }
    }
    return 0;
}


/* Ascii case-insensitive string comparison */
int
istrncmp(const char *x, const char *y, long len)
{
    const unsigned char *p = (unsigned char *)x;
    const unsigned char *q = (unsigned char *)y;

    while (len > 0) {
	if (tolower(*p) != tolower(*q)) {
	    return (*p - *q);
	} else if (*p == 0) {
	    break;
	}
	p++;
	q++;
    }
    return (0);
}
